from dataclasses import dataclass


@dataclass(frozen=True)
class FmtConf:
    fromat: str = ...
    asctime: bool = True
    created: bool = True
    exc_info: bool = True
    filename: bool = True
    funcName: bool = True
    levelname: bool = True
    levelno: bool = True
    lineno: bool = True
    message: bool = True
    module: bool = True
    msecs: bool = True
    msg: bool = True
    name: bool = True
    pathname: bool = True
    process: bool = True
    processName: bool = True
    relativeCreated: bool = True
    stack_info: bool = True
    thread: bool = True
    threadName: bool = True

    def get(self) -> str:
        if bool(self.fromat):
            return self.fromat

        else:
            return self.__createConfStr()

    def __createConfStr(self) -> str:
        loggerFormat = ''

        for key, value in self.__dict__.items():
            if self.__isValid(key, value):
                loggerFormat += '' + f'%({key})s'

        return loggerFormat

    @classmethod
    def __isValid(cls, key, value):
        if key == 'format':
            return None

        return key if bool(value) else None



from dataclasses import dataclass


@dataclass(frozen=True)
class ElasticSearchConf:
    hosts: list[str]
    user: str
    password: str
    indicator: str
    isActive: bool = True
    maxConnection: int = 10
    sniffOnStart: bool = False
    sniffOnConnectionFail: bool = False
    snifferTimeout: float = 60
    sniffTimeout: float = 10
    isCriticalActive: bool = True
    isErrorActive: bool = True
    isWarningActive: bool = True
    isInfoActive: bool = True
    isDebugActive: bool = False

    def get(self) -> dict:
        return self.__conf()

    def __conf(self) -> dict:
        confDict: dict = self.__confDict()
        any(confDict.update({key: value}) for key, value in self.__dict__.items())
        return confDict

    @classmethod
    def __confDict(cls) -> dict:
        return {
            "hosts": ...,
            "http_auth": ...,
            "max_connection": ...,
            "sniff_on_start": ...,
            "sniff_on_connection_fail": ...,
            "snifferTimeout": ...,
            "sniffTimeout": ...
        }



import copy
import logging
from queue import Queue

from ._conf import ElasticSearchConf
from ._formatter import ElasticSearchFormatter
from ._els_logger import ElasticSearchLogger
from ._index import IndexFormatter
from advancedLogger.contract import ITimeFormatter, IIndexFormatter


class ELSHandler:

    def __init__(self, els_conf: ElasticSearchConf,
                 time_formatter: ITimeFormatter,
                 els_index_formatter: IIndexFormatter = None):

        super().__init__()
        self.__elsConf = els_conf
        self.__elsQueue: Queue = ...
        self.__elsLogger: ElasticSearchLogger = ...
        self.__indexFormatter: IIndexFormatter = els_index_formatter
        self.__timeFormatter: ITimeFormatter = time_formatter
        self.__elsFormatter: ElasticSearchFormatter = ElasticSearchFormatter(time_formatter)
        self.__setupElastic(els_conf)

    def __setupElastic(self, els_conf: ElasticSearchConf):
        self.__elsQueue = Queue()
        self.__setupIndexFormatter()
        self.__elsLogger = ElasticSearchLogger(index_formatter=self.__indexFormatter,
                                               config=els_conf,
                                               queue=self.__elsQueue)

    def start(self):
        self.__elsLogger.start()

    def __setupIndexFormatter(self):
        if not self.__indexFormatter:
            self.__indexFormatter = IndexFormatter(self.__elsConf.indicator, self.__timeFormatter)

    def emit(self, record: logging.LogRecord):
        match record.levelname:
            case "CRITICAL":
                if self.__elsConf.isCriticalActive:
                    self.__emit(record)

            case "ERROR":
                if self.__elsConf.isErrorActive:
                    self.__emit(record)

            case "INFO":
                if self.__elsConf.isInfoActive:
                    self.__emit(record)

            case "DEBUG":
                if self.__elsConf.isDebugActive:
                    self.__emit(record)

            case "WARNING":
                if self.__elsConf.isWarningActive:
                    self.__emit(record)

            case "FATAL":
                if self.__elsConf.isCriticalActive:
                    self.__emit(record)

    def __emit(self, record: logging.LogRecord):
        elsLogRecord = self.__elsFormatter.format(record)
        self.__elsQueue.put(elsLogRecord)





import logging
from ._els_logger import ELSHandler, ElasticSearchConf
from .contract import IIndexFormatter, ITimeFormatter
from ._time import TimeFormatter


class AdvancedLogHandler(logging.StreamHandler):

    def __init__(self, els_conf: ElasticSearchConf = None,
                 els_index_formatter: IIndexFormatter = None,
                 time_formatter: ITimeFormatter = None):
        super(AdvancedLogHandler, self).__init__()
        self.__elsHandler = ...
        self.__elsConf: ElasticSearchConf = els_conf
        self.__isElsActive = False

        self.__elsEvaluation(els_index_formatter, time_formatter)

    def __elsEvaluation(self, els_index_formatter: IIndexFormatter, time_formatter: ITimeFormatter):
        if not time_formatter:
            time_formatter = TimeFormatter()

        if self.__elsConf:
            self.__elsHandler = ELSHandler(self.__elsConf, time_formatter, els_index_formatter)
            if self.__elsConf.isActive:
                self.__elsHandler.start()
                self.__isElsActive = True

    def emit(self, record: logging.LogRecord) -> None:
        if self.__isElsActive:
            self.__elsHandler.emit(record)
        super().emit(record)



from ._els_logger._index import IndexFormatter
from .contract import IIndexFormatter, ITimeFormatter
from ._handler import AdvancedLogHandler
from ._formatter import AdvancedLogFormatter
from ._formatter import FmtConf
from ._els_logger import ElasticSearchConf
from ._time import TimeFormatter


__title__ = "advancelogger"

__author__ = "Matin Karbasioun"

__publish__ = "2022"

__license__ = "MIT License"

__copyright__ = f"Copyright (c) {__publish__} " + __author__

from .contract import IIndexFormatter

ElsConfig: ElasticSearchConf = ...
DateFormatter = TimeFormatter()



import logging
import json
# from advancelogger._elasticsearch_logger import ElasticSearchConf, ElasticSearchLogger, IndexFormatter
# from ._fmt_conf import FmtConf


class AdvancedLogFormatter(logging.Formatter):
    """
        Use Advance Log Formatter to emphasize on how and which contents your want in your log

        Based on built-in logging library in python, Formatters responsible for converting
        a LogRecord to a string which be human-readable and can show important information
        on the log record

        Advance log formatter needs to know about which information and how order them for
        log record. This can define by the initial string call fmt (short stand for format) at
        initialize the logger in the python code.

         %(name)s            Name of the logger (logging channel)
        %(levelno)s         Numeric logging level for the message (DEBUG, INFO,
                            WARNING, ERROR, CRITICAL)
        %(levelname)s       Text logging level for the message ("DEBUG", "INFO",
                            "WARNING", "ERROR", "CRITICAL")
        %(pathname)s        Full pathname of the source file where the logging
                            call was issued (if available)
        %(filename)s        Filename portion of pathname
        %(module)s          Module (name portion of filename)
        %(lineno)d          Source line number where the logging call was issued
                            (if available)
        %(funcName)s        Function name
        %(created)f         Time when the LogRecord was created (time.time()
                            return value)
        %(asctime)s         Textual time when the LogRecord was created
        %(msecs)d           Millisecond portion of the creation time
        %(relativeCreated)d Time in milliseconds when the LogRecord was created,
                            relative to the time the logging module was loaded
                            (typically at application startup time)
        %(thread)d          Thread ID (if available)
        %(threadName)s      Thread name (if available)
        %(process)d         Process ID (if available)
        %(message)s         The result of record.getMessage(), computed just as
                            the record is emitted

    """

    def __init__(self, log_fmt: str | None = None, date_fmt: str | None = None, index_splitter: str | None = '%'):
        super().__init__(fmt=log_fmt, datefmt=date_fmt, style=index_splitter)
        self.__isJsonActive = False

    def useJson(self):
        self.__isJsonActive = True

    def formatMessage(self, record) -> dict:
        """
        Overwritten to return a dictionary of the relevant LogRecord attributes instead of a string.
        KeyError is raised if an unknown attribute is provided in the fmt_dict.
        """

        return record.__dict__

    def format(self, record) -> str:
        """
        Mostly the same as the parent's class method, the difference being that a dict is manipulated and dumped as JSON
        instead of a string.
        """
        record.message = record.getMessage()

        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)

        message_dict = self.formatMessage(record)

        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)

        if record.exc_text:
            message_dict["exc_info"] = record.exc_text

        if record.stack_info:
            message_dict["stack_info"] = self.formatStack(record.stack_info)

        if self.__isJsonActive:
            return json.dumps(message_dict, default=str)

        else:
            return record.message



import logging
import time
from queue import Queue
from threading import Thread
import urllib3
from elasticsearch import Elasticsearch
from ._conf import ElasticSearchConf
from advancedLogger.contract import IIndexFormatter

logger = logging.getLogger(__name__)


class ElasticSearchLogger(Thread):
    def __init__(self, index_formatter: IIndexFormatter, config: ElasticSearchConf, queue: Queue):
        super().__init__()

        self.__queue = queue
        self.__config: ElasticSearchConf = config
        self.__indexFormatter: IIndexFormatter = index_formatter

        self.__es = Elasticsearch(hosts=self.__config.hosts,
                                  http_auth=(self.__config.user, self.__config.password),
                                  sniff_on_start=self.__config.sniffOnStart,
                                  sniff_on_connection_fail=self.__config.sniffOnConnectionFail,
                                  sniffer_timeout=self.__config.snifferTimeout,
                                  sniff_timeout=self.__config.sniffTimeout)

    def run(self):

        while True:

            try:
                if not self.__queue.empty():
                    body = self.__queue.get()
                    self.__es.index(index=self.__indexFormatter.getIndex(), document=body)

            except urllib3.exceptions.ConnectTimeoutError:
                logger.error('Elastic Search Connection Error')

            except Exception as e:
                logger.exception(e)

            time.sleep(0.01)



from advancedLogger.contract import IIndexFormatter, ITimeFormatter


class IndexFormatter(IIndexFormatter):
    def __init__(self, indicator: str, timeFormatter: ITimeFormatter):
        self.__timeFormatter: ITimeFormatter = timeFormatter
        self.__indicator = indicator

    def getIndex(self) -> str:

        return self.__indicator + '-' + self.__timeFormatter.get()



import copy
import logging
import uuid

from advancedLogger.contract import ITimeFormatter


class ElasticSearchFormatter:
    def __init__(self, time_formatter: ITimeFormatter):
        self.__time_formatter: ITimeFormatter = time_formatter

    def format(self, logRecord: logging.LogRecord) -> dict:
        logDict = copy.deepcopy(logRecord.__dict__)
        logDict.pop("args")
        logDict.update({"@timestamp": self.__time_formatter.convert(logRecord.created)})
        logDict.update({"traceId": str(uuid.uuid4())})
        return logDict



from datetime import datetime
from typing import Optional

import pytz


class TimeFormatter:
    def __init__(self, date_fmt: Optional[str] = "%Y-%m-%d", tz: Optional[pytz.BaseTzInfo] = pytz.UTC):
        self.__tz = tz
        self.__dt_fmt = date_fmt

    def get(self):
        return datetime.today().strftime(self.__dt_fmt)

    def convert(self, time: int):
        return datetime.fromtimestamp(time, self.__tz).strftime(self.__dt_fmt)
